use super::*;

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Point<'a, Ctx, A> {
    /// Make a [Point] from an [Address].
    pub fn from_address(
        address: Address,
        factory: A::Fctr,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> Self {
        Point {
            point: address.point,
            origin: Rc::new(ResolverOrigin {
                r_factory: factory,
                r_address: address,
                r_resolver: resolver,
            }),
        }
    }
}

struct ResolverOrigin<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> {
    r_factory: F,
    r_address: Address,
    r_resolver: Rc<dyn Resolver<'a, Ctx>>,
}

fn _resolve_origin<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
    origin: Rc<ResolverOrigin<'a, Ctx, A::Fctr>>,
) -> Resolution<'a, Ctx, A> {
    let resolution = origin.r_resolver.clone().resolve(origin.r_address);
    Ctx::T::fmap(
        move |resolved| match resolved {
            Ok((src, resolver)) => match origin.r_factory.parse_slice(&src, resolver) {
                Ok(mentionable) => Ok(Rc::new(mentionable)),
                Err(parse_error) => Err(ResolutionError::Parse(parse_error)),
            },
            Err(lookup_error) => Err(ResolutionError::Lookup(lookup_error)),
        },
        resolution,
    )
}

impl<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> Origin<'a, Ctx> for ResolverOrigin<'a, Ctx, F> {
    type Mtbl = F::Mtbl;

    fn factory(&self) -> <Self::Mtbl as Mentionable<'a, Ctx>>::Fctr {
        self.r_factory.clone()
    }

    fn resolve(self: Rc<Self>) -> Resolution<'a, Ctx, Self::Mtbl> {
        _resolve_origin(self)
    }
}
