use std::cmp::min;

use super::*;

/// [Deserializer] for slices. Used in [`ExtFactory::parse_slice`] and [`crate::std::atomic::ExtAtomic::parse_slice`].
pub struct SliceDeserializer<'a> {
    slice: &'a [u8],
    pos: usize,
}

impl<'a> From<&'a [u8]> for SliceDeserializer<'a> {
    fn from(value: &'a [u8]) -> Self {
        SliceDeserializer {
            slice: value,
            pos: 0,
        }
    }
}

impl<'a> Deserializer for SliceDeserializer<'a> {
    fn read_n(&mut self, n: usize) -> &[u8] {
        let (left, right) = self.slice.split_at(min(n, self.slice.len()));
        self.slice = right;
        self.pos += left.len();
        left
    }

    fn read_all(&mut self) -> &[u8] {
        let left = self.slice;
        self.slice = &[];
        self.pos += left.len();
        left
    }

    fn tell(&self) -> usize {
        self.pos
    }
}
